package org.jbpm.possiblebug;

import java.util.Collection;
import java.util.HashMap;
import java.util.Properties;

import javax.persistence.Persistence;

import org.drools.core.impl.EnvironmentFactory;
import org.drools.core.io.impl.ClassPathResource;
import org.jbpm.workflow.instance.NodeInstance;
import org.jbpm.workflow.instance.impl.WorkflowProcessInstanceImpl;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.kie.api.KieBase;
import org.kie.api.KieServices;
import org.kie.api.builder.KieBuilder;
import org.kie.api.builder.KieFileSystem;
import org.kie.api.builder.KieRepository;
import org.kie.api.builder.Message.Level;
import org.kie.api.persistence.jpa.KieStoreServices;
import org.kie.api.runtime.Environment;
import org.kie.api.runtime.EnvironmentName;
import org.kie.api.runtime.KieContainer;
import org.kie.api.runtime.KieSession;
import org.kie.api.runtime.KieSessionConfiguration;
import org.kie.api.runtime.process.ProcessInstance;
import org.kie.api.runtime.process.WorkItem;
import org.kie.api.runtime.process.WorkItemHandler;
import org.kie.api.runtime.process.WorkItemManager;
import org.kie.internal.KnowledgeBaseFactory;

import bitronix.tm.TransactionManagerServices;
import bitronix.tm.resource.jdbc.PoolingDataSource;

public class PossibleBugTest {

	private PoolingDataSource ds = new PoolingDataSource();
	
	@Before
	public void setUp() throws Exception {
		
		 ds.setUniqueName("jdbc/testDS1");
	     //NON XA CONFIGS
	     ds.setClassName("org.h2.jdbcx.JdbcDataSource");
	     ds.setMaxPoolSize(3);
	     ds.setAllowLocalTransactions(true);
	     ds.getDriverProperties().put("user", "sa");
	     ds.getDriverProperties().put("password", "sasa");
	     ds.getDriverProperties().put("URL", "jdbc:h2:mem:mydb");
	     ds.init();
	}
	
	@After
	public void tearDown() throws Exception {
		TransactionManagerServices.getTransactionManager().shutdown();
		ds.close();
		System.gc();
	}
	
	@Test
	public void testTimerAndGateway() throws Exception {
		KieServices ks = KieServices.Factory.get();
		KieBase kbase = generateKieBase(ks, "processes/timer-process.bpmn2");
		Properties ksprops = new Properties();
		KieSessionConfiguration ksconf = KnowledgeBaseFactory.newKnowledgeSessionConfiguration(ksprops);
		Environment env = createEnvironment();
		KieStoreServices kstore = ks.getStoreServices();
		KieSession ksession = kstore.newKieSession(kbase, ksconf, env);
		int sessionId = ksession.getId();
		
		TestAsyncWorkItemHandler handler1 = new TestAsyncWorkItemHandler();
		TestAsyncWorkItemHandler handler2 = new TestAsyncWorkItemHandler();
		
		ksession.getWorkItemManager().registerWorkItemHandler("task1", handler1);
		ksession.getWorkItemManager().registerWorkItemHandler("task2", handler2);

		ProcessInstance instance = ksession.createProcessInstance("timer-process", new HashMap<String, Object>());
		ksession.startProcessInstance(instance.getId());

		WorkItem workItem1 = handler1.getWorkItem();
		Assert.assertNotNull(workItem1);
		Assert.assertNull(handler1.getWorkItem());
		//first safe state: task1 completed
		ksession.getWorkItemManager().completeWorkItem(workItem1.getId(), null);
		ksession.dispose();
		
		ksession = kstore.loadKieSession(sessionId, kbase, ksconf, env);
		//second safe state: timer completed, waiting on task2
		Thread.sleep(5000);

		WorkItem workItem2 = handler2.getWorkItem();
                //Both sides of the join are completed. But on the process instance, there are two
                //JoinInstance for the same Join, and since it is an AND join, it never reaches task2
                //It fails after the next assertion
		Assert.assertNotNull(workItem2);
		Assert.assertNull(handler2.getWorkItem());
	}

	private Environment createEnvironment() {
		Environment env = EnvironmentFactory.newEnvironment();
		env.set(EnvironmentName.ENTITY_MANAGER_FACTORY, Persistence.createEntityManagerFactory("org.jbpm.persistence.jpa"));
		env.set(EnvironmentName.TRANSACTION_MANAGER, TransactionManagerServices.getTransactionManager());
		return env;
	}

	private KieBase generateKieBase(KieServices ks, String processFile) {
		KieRepository kr = ks.getRepository();
		ks.getKieClasspathContainer();
		KieFileSystem kfs = ks.newKieFileSystem();
		kfs.write(new ClassPathResource(processFile));
		KieBuilder kbuilder = ks.newKieBuilder(kfs);
		kbuilder.buildAll();
		if (kbuilder.getResults().hasMessages(Level.ERROR)) {
			throw new IllegalStateException("Problem reading bpmn2: " + kbuilder.getResults().toString());
		}
		KieContainer kcont = ks.newKieContainer(kr.getDefaultReleaseId());
		return kcont.getKieBase();
	}

	private static class TestAsyncWorkItemHandler implements WorkItemHandler {

		private WorkItem workItem;
		private int activations = 0;
		
		public void executeWorkItem(WorkItem workItem, WorkItemManager manager) {
			System.out.println("Starting call to handler " + workItem.getName());
			this.workItem = workItem;
			this.activations++;
		}

		public void abortWorkItem(WorkItem workItem, WorkItemManager manager) {
			this.workItem = null;
		}
		
		public WorkItem getWorkItem() {
			WorkItem result = workItem;
			workItem = null;
			return result;
		}
		
		public int getActivations() {
			return activations;
		}
	}
}

